@@ -27,6 +27,94 @@
  * Contributor(s): ______________________________________.
  *
  * $Log: ixjunix.cxx,v $
+ * Revision 1.143  2004/02/06 10:19:57  csoutheren
+ * Added PThread::Yield to ReadFrame and WriteFrame calls to see if this
+ * fixes the starvation problem on 2.6 kernels
+ *
+ * Revision 1.142  2004/01/31 13:13:22  csoutheren
+ * Fixed problem with HAS_IXJ being tested but not included
+ *
+ * Revision 1.141  2004/01/09 10:13:33  dereksmithies
+ * Reads hookstate at startup, and does not assume phone is on hook.
+ *
+ * Revision 1.140  2003/12/23 22:52:56  dsandras
+ * IXJ devices that fail opening because they are busy should be listed in the list of detected devices.
+ *
+ * Revision 1.139  2003/11/10 12:04:25  dsandras
+ * Added missing math.h include.
+ *
+ * Revision 1.138  2003/10/24 03:39:54  dereksmithies
+ * Add card dependant dsp maximums. and log scaling volumes.
+ *
+ * Revision 1.137  2003/10/17 01:33:24  dereksmithies
+ * Code now correctly reports the type string of the card.
+ *
+ * Revision 1.136  2003/04/29 08:31:44  robertj
+ * Fixed return type of get wink function.
+ *
+ * Revision 1.135  2003/04/28 01:47:52  dereks
+ * Add ability to set/get wink duration for ixj device.
+ *
+ * Revision 1.134  2002/11/05 04:44:20  robertj
+ * Fixed typo
+ *
+ * Revision 1.133  2002/11/05 04:38:07  robertj
+ * Changed IsLineDisconnected() to work with POTSLine
+ *
+ * Revision 1.132  2002/11/05 04:27:34  robertj
+ * Imported RingLine() by array from OPAL.
+ *
+ * Revision 1.131  2002/10/24 21:06:28  dereks
+ * Add additional PTRACE statements to aid in debugging.
+ *
+ * Revision 1.130  2002/10/01 06:43:01  robertj
+ * Removed GNU compiler warning
+ *
+ * Revision 1.129  2002/08/30 08:20:22  craigs
+ * Added G.723.1A based codecs
+ *
+ * Revision 1.128  2002/08/05 10:03:47  robertj
+ * Cosmetic changes to normalise the usage of pragma interface/implementation.
+ *
+ * Revision 1.127  2002/07/25 09:57:57  rogerh
+ * Make sure we pick the G.723.1 6.3k codec before the 5.3k codec
+ * if it is available.
+ *
+ * Revision 1.126  2002/07/19 10:47:16  robertj
+ * Fixed bug where can still receive 24 byte frames even whan selected
+ *   G.723.1(5.3k) mode, it only controls what is transmitted.
+ *
+ * Revision 1.125  2002/06/25 09:56:07  robertj
+ * Fixed GNU warnings
+ *
+ * Revision 1.124  2002/05/21 10:21:39  robertj
+ * Fixed FLASH_TIME being correctly set, PHONE_RING_START getting correct
+ *   argument, correct setting of ancestor variables when stopping codec,
+ *   odd packet error if playout time passed, all thanks Artis Kugevics
+ *
+ * Revision 1.123  2002/05/09 06:26:34  robertj
+ * Added fuction to get the current audio enable state for line in device.
+ * Changed IxJ EnableAudio() semantics so is exclusive, no direct switching
+ *   from PSTN to POTS and vice versa without disabling the old one first.
+ *
+ * Revision 1.122  2002/02/08 14:38:42  craigs
+ * Changed codec table to use defines from mediafmt.h. Thanks to Roger Hardiman
+ *
+ * Revision 1.121  2002/02/06 06:51:17  rogerh
+ * Correct a comment. There are 240 samples in PCM16 audio (not 250)
+ *
+ * Revision 1.120  2002/01/23 19:09:45  rogerh
+ * Return PCM as a supported format type
+ *
+ * Revision 1.119  2001/12/19 07:42:52  craigs
+ * Fixed problem with opening devices as returned by GetDeviceName
+ *
+ * Revision 1.118  2001/12/12 21:54:33  craigs
+ * Changed for changes for G723.1 5k3
+ *
+ * Revision 1.117  2001/11/22 02:55:13  robertj
+ * Fixed bug from previous fix, caused only /dev/phone0 to work!
+ *
  * Revision 1.116  2001/11/09 07:48:01  craigs
  * Changed to allow Open to accept string returned by GetName
  *
@@ -394,15 +482,18 @@
  *
  */
 
+#include <ptlib.h>
+
 #ifdef __GNUC__
 #pragma implementation "ixjlid.h"
 #endif
 
-
-#include <ptlib.h>
 #include "ixjlid.h"
 
+#ifdef HAS_IXJ
+
 #include <sys/time.h>
+#include <math.h>
 
 #define new PNEW
 
@@ -420,7 +511,7 @@
 #endif
 
 
-#define	IsLineJACK()	(dwCardType==3)
+#define	IsLineJACK()	(dwCardType == LineJACK)
 
 #define	FLASH_TIME	1000
 #define	MANUAL_FLASH		// undefine to use FLASH exception
@@ -663,6 +754,10 @@ OpalIxJDevice::OpalIxJDevice()
   readStopped = writeStopped = TRUE;
   readFrameSize = writeFrameSize = 480;  // 30 milliseconds of 16 bit PCM data
   readCodecType = writeCodecType = P_MAX_INDEX;
+  currentHookState = lastHookState = FALSE;
+  inRawMode = FALSE;
+  enabledAudioLine = UINT_MAX;
+  exclusiveAudioMode = TRUE;
   aecLevel = AECOff;
   tonePlaying = FALSE;
   removeDTMF = FALSE;
@@ -687,19 +782,21 @@ BOOL OpalIxJDevice::Open(const PString & device)
     }
   }
 
-  if (!isdigit(device[0])) 
+  if (isdigit(device[0])) 
     deviceName = psprintf("/dev/phone%u", device.AsUnsigned());
   else {
-    PINDEX pos = deviceName.FindLast(' ');
+    PINDEX pos = device.FindLast(' ');
     if (pos == P_MAX_INDEX)
       deviceName = device;
     else
-      deviceName = device.Left(pos);
+      deviceName = device.Mid(pos+1).Trim();
   }
 
   int new_handle = os_handle = ::open(deviceName, O_RDWR);
   if (!ConvertOSError(new_handle))
     return FALSE;
+  currentHookState = lastHookState = 
+    (IOCTL(os_handle, PHONE_HOOKSTATE) != 0);
 
   // add the new handle to the exception info
   {
@@ -714,7 +811,7 @@ BOOL OpalIxJDevice::Open(const PString & device)
     memset(&info, 0, sizeof(info));
 
     info.fd  = os_handle;
-    info.hookState  = FALSE;
+    info.hookState  = currentHookState;
     info.hasRing    = FALSE;
     info.hasWink    = FALSE;
     info.hasFlash   = FALSE;
@@ -767,7 +864,31 @@ BOOL OpalIxJDevice::Open(const PString & device)
   os_handle = new_handle;
 
   // determine if the card is a phonejack or linejack
-  dwCardType = IOCTL(os_handle, IXJCTL_CARDTYPE)/100;
+  dwCardType = IOCTL(os_handle, IXJCTL_CARDTYPE);
+  switch (dwCardType) {
+  case 3:
+    dwCardType = PhoneJACK_PCI_TJ;
+    break;
+  case 0x100: 
+  case 100:
+    dwCardType = PhoneJACK;
+    break;
+  case 0x300:
+  case 300:
+    dwCardType = LineJACK;
+    break;
+  case 0x400:
+  case 400:
+    dwCardType = PhoneJACK_Lite;
+    break;
+  case 0x500:
+  case 500: 
+    dwCardType = PhoneJACK_PCI;
+    break;
+  case 0x600:
+  case 600:
+    dwCardType = PhoneCARD;
+  }
 
   char * str = ::getenv("IXJ_COUNTRY");
   if (str != NULL) {
@@ -782,8 +903,6 @@ BOOL OpalIxJDevice::Open(const PString & device)
   gotWink       = FALSE;
   IOCTL2(os_handle, PHONE_PSTN_SET_STATE, PSTN_ON_HOOK);
 
-  lastHookState = FALSE;
-
   inRawMode     = FALSE;
 
   SetAEC         (0, AECOff);
@@ -823,29 +942,34 @@ BOOL OpalIxJDevice::Close()
 PString OpalIxJDevice::GetName() const
 {
   switch (dwCardType) {
-    case 0:
-    case 1:
-      return "Internet PhoneJACK-ISA " + deviceName;
-
-    case 3:
-      return "Internet LineJACK " + deviceName;
-
-    case 4:
-      return "Internet PhoneJACK-Lite " + deviceName;
-
-    case 5:
-      return "Internet PhoneJACK-PCI " + deviceName;
-
-    case 6:
-      return "Internet PhoneCARD " + deviceName;
+  case 0:
+  case PhoneJACK:
+    return "Internet PhoneJACK-ISA " + deviceName;
+    
+  case LineJACK:
+    return "Internet LineJACK " + deviceName;
+    
+  case PhoneJACK_Lite:
+    return "Internet PhoneJACK-Lite " + deviceName;
+    
+  case PhoneJACK_PCI:
+    return "Internet PhoneJACK-PCI " + deviceName;
+    
+  case PhoneCARD:
+    return "Internet PhoneCARD " + deviceName;
+    
+  case PhoneJACK_PCI_TJ:
+    return "Internet PhoneJack-PCI " + deviceName;
   }
-
+  
   return "xJACK " + deviceName;
 }
 
 unsigned OpalIxJDevice::GetLineCount()
 {
-  return IsLineJACK() ? NumLines : 1;
+  return 1;     
+  /*Each card has just one line that can be active. 
+    Consequently, return (IsLineJACK() ? NumLines : 1); is wrong.*/
 }
 
 BOOL OpalIxJDevice::IsLinePresent(unsigned line, BOOL /* force */)
@@ -854,7 +978,7 @@ BOOL OpalIxJDevice::IsLinePresent(unsigned line, BOOL /* force */)
     return FALSE;
 
   BOOL stat = IOCTL(os_handle, IXJCTL_PSTN_LINETEST) == 1;
-  PThread::Current()->Sleep(2000);
+  PThread::Sleep(2000);
 
   // clear ring signal status
   IsLineRinging(line);
@@ -871,13 +995,21 @@ BOOL OpalIxJDevice::IsLineOffHook(unsigned line)
   PWaitAndSignal m(exceptionMutex);
   ExceptionInfo * info = GetException();
 
+#ifdef MANUAL_FLASH
   if (info->hookState != lastHookState) {
     lastHookState = info->hookState;
-    hookTimeout = FLASH_TIME;
+    if (lastHookState) {
+	currentHookState = lastHookState;
+    } else {
+	hookTimeout = FLASH_TIME;
+    }
   } else if (!hookTimeout.IsRunning() && (currentHookState != info->hookState)) 
     currentHookState = info->hookState;
 
   return currentHookState;
+#else
+  return info->hookState;
+#endif
 }
 
 BOOL OpalIxJDevice::HasHookFlash(unsigned line)
@@ -948,7 +1080,7 @@ BOOL OpalIxJDevice::IsLineRinging(unsigned line, DWORD * /*cadence*/)
 
 BOOL OpalIxJDevice::RingLine(unsigned line, DWORD cadence)
 {
-  if (line != 0)
+  if (line != POTSLine)
     return FALSE;
 
   if (cadence == 0)
@@ -966,17 +1098,31 @@ BOOL OpalIxJDevice::RingLine(unsigned line, DWORD cadence)
     SetCallerID(line, "");
   } else
 #endif
-    stat = IOCTL(os_handle, PHONE_RING_START);
+    stat = IOCTL2(os_handle, PHONE_RING_START, 0);
 
   return ConvertOSError(stat);
 }
 
 
+BOOL OpalIxJDevice::RingLine(unsigned line, PINDEX nCadence, unsigned * pattern)
+{
+  if (line >= GetLineCount())
+    return FALSE;
+
+  if (line != POTSLine)
+    return FALSE;
+
+  return RingLine(line, nCadence != 0 ? 0xaaa : 0);
+}
+
+
 BOOL OpalIxJDevice::IsLineDisconnected(unsigned line, BOOL checkForWink)
 {
+  if (line >= GetLineCount())
+    return FALSE;
 
   if (line != PSTNLine)
-    return FALSE;
+    return !IsLineOffHook(line);
 
   if (checkForWink) {
 
@@ -1030,17 +1176,23 @@ BOOL OpalIxJDevice::ConvertOSError(int err)
 
 static const struct {
   const char * mediaFormat;
-  PINDEX frameSize;
+  PINDEX writeFrameSize;
+  PINDEX readFrameSize;
   int mode;
   int frameTime;
+  BOOL vad;
 } CodecInfo[] = {
-  { OPAL_PCM16,       480, LINEAR16, 30 },   // 480 bytes = 250 samples = 30ms
-  { "G.711-uLaw-64k", 240, ULAW,     30 },   // 240 bytes = 240 samples = 30ms
-  { "G.711-ALaw-64k", 240, ALAW,     30 },   // 240 bytes = 240 samples = 30ms
-  { "G.728",           60, G728,     30 },   // 60 bytes  = 12 frames   = 30ms
-  { "G.729A",          10, G729,     10 },   // 10 bytes = 1 frame = 10 ms
-  { "G.729A/B",        10, G729B,    10 },   // 10 bytes = 1 frame = 10 ms
-  { "G.723.1",         24, G723_63,  30 }    // 24 bytes = 1 frame = 30 ms
+  /* NOTE: These are enumerated in reverse order. */
+  { OPAL_PCM16,         480, 480, LINEAR16, 30, FALSE },   // 480 bytes = 240 samples = 30ms
+  { OPAL_G711_ULAW_64K, 240, 240, ULAW,     30, FALSE },   // 240 bytes = 240 samples = 30ms
+  { OPAL_G711_ALAW_64K, 240, 240, ALAW,     30, FALSE },   // 240 bytes = 240 samples = 30ms
+  { OPAL_G728,           60,  60, G728,     30, FALSE },   // 60 bytes  = 12 frames   = 30ms
+  { OPAL_G729A,          10,  10, G729,     10, FALSE },   // 10 bytes = 1 frame = 10 ms
+  { OPAL_G729AB,         10,  10, G729B,    10,  TRUE },   // 10 bytes = 1 frame = 10 ms
+  { OPAL_G7231_5k3 ,     24,  20, G723_53,  30, FALSE },   // 20 bytes = 1 frame = 30 ms
+  { OPAL_G7231_6k3,      24,  24, G723_63,  30, FALSE },   // 24 bytes = 1 frame = 30 ms
+  { OPAL_G7231A_5k3 ,    24,  20, G723_53,  30,  TRUE },   // 20 bytes = 1 frame = 30 ms
+  { OPAL_G7231A_6k3,     24,  24, G723_63,  30,  TRUE }    // 24 bytes = 1 frame = 30 ms
 };
 
 
@@ -1050,7 +1202,7 @@ OpalMediaFormat::List OpalIxJDevice::GetMediaFormats() const
   OpalMediaFormat::List codecs;
 
   PINDEX idx = PARRAYSIZE(CodecInfo);
-  while (idx-- > 1) { // Stop before we get to PCM
+  while (idx-- > 0) {
     phone_capability cap;
     cap.captype = codec;
     cap.cap = CodecInfo[idx].mode;
@@ -1088,6 +1240,7 @@ BOOL OpalIxJDevice::SetReadFormat(unsigned line, const OpalMediaFormat & mediaFo
   if (!readStopped) {
     IOCTL(os_handle, PHONE_REC_STOP);
     readStopped = TRUE;
+    OpalLineInterfaceDevice::StopReadCodec(line);
   }
 
   readCodecType = FindCodec(mediaFormat);
@@ -1107,7 +1260,7 @@ BOOL OpalIxJDevice::SetReadFormat(unsigned line, const OpalMediaFormat & mediaFo
          << CodecInfo[readCodecType].mediaFormat
          << " code=" << CodecInfo[readCodecType].mode);
 
-  readFrameSize = CodecInfo[readCodecType].frameSize;
+  readFrameSize = CodecInfo[readCodecType].readFrameSize;
 
   // set frame time
   if (writeStopped)
@@ -1152,6 +1305,7 @@ BOOL OpalIxJDevice::SetWriteFormat(unsigned line, const OpalMediaFormat & mediaF
   if (!writeStopped) {
     IOCTL(os_handle, PHONE_PLAY_STOP);
     writeStopped = TRUE;
+    OpalLineInterfaceDevice::StopWriteCodec(line);
   }
 
 
@@ -1172,7 +1326,7 @@ BOOL OpalIxJDevice::SetWriteFormat(unsigned line, const OpalMediaFormat & mediaF
          << CodecInfo[writeCodecType].mediaFormat
          << " code=" << CodecInfo[writeCodecType].mode);
 
-  writeFrameSize = CodecInfo[writeCodecType].frameSize;
+  writeFrameSize = CodecInfo[writeCodecType].writeFrameSize;
 
   // set frame time
   if (readStopped)
@@ -1260,6 +1414,7 @@ BOOL OpalIxJDevice::SetRawCodec(unsigned line)
 
   if (!SetReadFormat (line, CodecInfo[0].mediaFormat) ||
       !SetWriteFormat(line, CodecInfo[0].mediaFormat)) {
+    PTRACE(1, "IXJ\t Failed to set raw codec");
     StopReadCodec(line);
     StopWriteCodec(line);
     return FALSE;
@@ -1267,8 +1422,8 @@ BOOL OpalIxJDevice::SetRawCodec(unsigned line)
 
   // set the new (maximum) volumes
   SetAEC         (line, AECOff);
-  SetRecordVolume(line, 0x100);
-  SetPlayVolume  (line, 0x100);
+  SetRecordVolume(line, 100);
+  SetPlayVolume  (line, 100);
 
   // stop values from changing
   inRawMode = TRUE;
@@ -1352,119 +1507,125 @@ static const PINDEX G723count[4] = { 24, 20, 4, 1 };
 
 BOOL OpalIxJDevice::ReadFrame(unsigned, void * buffer, PINDEX & count)
 {
-  PWaitAndSignal rmutex(readMutex);
+  {
+    PWaitAndSignal rmutex(readMutex);
 
-  count = 0;
+    count = 0;
 
-  if (readStopped) 
-    return FALSE;
+    if (readStopped) {
+        PTRACE(1, "IXJ\tRead stopped, so ReadFrame returns false");    
+        return FALSE;
+    }
 
-  if (writeStopped) {
-    PThread::Current()->Sleep(30);
-    memset(buffer, 0, readFrameSize);
+    if (writeStopped) {
+      PThread::Sleep(30);
+      memset(buffer, 0, readFrameSize);
+      switch (CodecInfo[readCodecType].mode) {
+        case G723_63:
+        case G723_53:
+          *((DWORD *)buffer) = 0x02;
+          count = 4;
+          break;
+        case G729B:
+          *((WORD *)buffer) = 0;
+          count = 2;
+          break;
+        default:
+          memset(buffer, 0, readFrameSize);
+          count = readFrameSize;
+          break;
+      }
+      return TRUE;
+    }
+  
+    WORD temp_frame_buffer[48];   // 30ms = 12 frames = 48 vectors = 48 WORDS for unpacked vectors
+    void * readBuf;
+    int    readLen;
+    switch (CodecInfo[readCodecType].mode) {
+      case G728 :
+        readBuf = temp_frame_buffer;
+        readLen = sizeof(temp_frame_buffer);
+        break;
+      case G729B :
+        readBuf = temp_frame_buffer;
+        readLen = 12;
+        break;
+      default :
+        readBuf = buffer;
+        readLen = readFrameSize;
+    }
+  
+    for (;;) {
+      fd_set rfds;
+      FD_ZERO(&rfds);
+      FD_SET(os_handle, &rfds);
+      struct timeval ts;
+      ts.tv_sec = 30;
+      ts.tv_usec = 0;
+  #if PTRACING
+      PTime then;
+  #endif
+      int stat = ::select(os_handle+1, &rfds, NULL, NULL, &ts);
+      if (stat == 0) {
+        PTRACE(1, "IXJ\tRead timeout:" << (PTime() - then));
+        return FALSE;
+      }
+      
+      if (stat > 0) {
+        stat = ::read(os_handle, readBuf, readLen);
+        if (stat == (int)readLen)
+          break;
+      }
+  
+      if ((stat >= 0) || (errno != EINTR)) {
+        PTRACE(1, "IXJ\tRead error = " << errno);
+        return FALSE;
+      }
+  
+      PTRACE(1, "IXJ\tRead EINTR");
+    }
+  
     switch (CodecInfo[readCodecType].mode) {
       case G723_63:
       case G723_53:
-        *((DWORD *)buffer) = 0x02;
-        count = 4;
+        count = G723count[(*(BYTE *)readBuf)&3];
         break;
-      case G729B:
-        *((WORD *)buffer) = 0;
-        count = 2;
-        break;
-      default:
-        memset(buffer, 0, readFrameSize);
+  
+      case G728 :
+        // for G728, pack four x 4 vectors
+        PINDEX i;
+        for (i = 0; i < 12; i++)
+          G728_Pack(temp_frame_buffer+i*4, ((BYTE *)buffer)+i*5);
         count = readFrameSize;
         break;
-    }
-    return TRUE;
-  }
-
-  WORD temp_frame_buffer[48];   // 30ms = 12 frames = 48 vectors = 48 WORDS for unpacked vectors
-  void * readBuf;
-  int    readLen;
-  switch (CodecInfo[readCodecType].mode) {
-    case G728 :
-      readBuf = temp_frame_buffer;
-      readLen = sizeof(temp_frame_buffer);
-      break;
-    case G729B :
-      readBuf = temp_frame_buffer;
-      readLen = 12;
-      break;
-    default :
-      readBuf = buffer;
-      readLen = readFrameSize;
-  }
-
-  for (;;) {
-    fd_set rfds;
-    FD_ZERO(&rfds);
-    FD_SET(os_handle, &rfds);
-    struct timeval ts;
-    ts.tv_sec = 30;
-    ts.tv_usec = 0;
-#if PTRACING
-    PTime then;
-#endif
-    int stat = ::select(os_handle+1, &rfds, NULL, NULL, &ts);
-    if (stat == 0) {
-      PTRACE(1, "IXJ\tRead timeout:" << (PTime() - then));
-      return FALSE;
-    }
-    
-    if (stat > 0) {
-      stat = ::read(os_handle, readBuf, readLen);
-      if (stat == (int)readLen)
+  
+      case G729B :
+        switch (temp_frame_buffer[0]) {
+          case 0 : // Silence
+            memset(buffer, 0, 10);
+            count = 10;
+            break;
+          case 1 : // Signal
+            memcpy(buffer, &temp_frame_buffer[1], 10);
+            count = 10;
+            break;
+          case 2 : // VAD
+            memcpy(buffer, &temp_frame_buffer[1], 2);
+            count = 2;
+            break;
+          default : // error
+            PTRACE(1, "IXJ\tIllegal value from codec in G729");
+            return FALSE;
+        }
         break;
+  
+      default :
+        count = readFrameSize;
     }
-
-    if ((stat >= 0) || (errno != EINTR)) {
-      PTRACE(1, "IXJ\tRead error = " << errno);
-      return FALSE;
-    }
-
-    PTRACE(1, "IXJ\tRead EINTR");
-  }
-
-  switch (CodecInfo[readCodecType].mode) {
-    case G723_63:
-    case G723_53:
-      count = G723count[(*(BYTE *)readBuf)&3];
-      break;
-
-    case G728 :
-      // for G728, pack four x 4 vectors
-      PINDEX i;
-      for (i = 0; i < 12; i++)
-        G728_Pack(temp_frame_buffer+i*4, ((BYTE *)buffer)+i*5);
-      count = readFrameSize;
-      break;
-
-    case G729B :
-      switch (temp_frame_buffer[0]) {
-        case 0 : // Silence
-          memset(buffer, 0, 10);
-          count = 10;
-          break;
-        case 1 : // Signal
-          memcpy(buffer, &temp_frame_buffer[1], 10);
-          count = 10;
-          break;
-        case 2 : // VAD
-          memcpy(buffer, &temp_frame_buffer[1], 2);
-          count = 2;
-          break;
-        default : // error
-          PTRACE(1, "IXJ\tIllegal value from codec in G729");
-          return FALSE;
-      }
-      break;
-
-    default :
-      count = readFrameSize;
   }
 
+  PThread::Yield();
+  
   return TRUE;
 }
 
@@ -1488,104 +1649,109 @@ static void G728_Unpack(const BYTE * packed, unsigned short * unpacked)
 
 BOOL OpalIxJDevice::WriteFrame(unsigned, const void * buffer, PINDEX count, PINDEX & written)
 {
-  PWaitAndSignal rmutex(readMutex);
-
-  written = 0;
-
-  if (writeStopped) 
-    return FALSE;
-
-  if (readStopped) {
-    PThread::Current()->Sleep(30);
-    return TRUE;
-  }
-
-  WORD temp_frame_buffer[48];
-  const void * writeBuf;
-  int writeLen;
-
-  switch (CodecInfo[writeCodecType].mode) {
-    case G723_63:
-    case G723_53:
-      writeBuf = buffer;
-      writeLen = 24;
-      written = G723count[(*(BYTE *)buffer)&3];
-      break;
-
-    case G728 :
-      writeBuf = temp_frame_buffer;
-      writeLen = sizeof(temp_frame_buffer);
-
-      // for G728, unpack twelve x 4 vectors
-      PINDEX i;
-      for (i = 0; i < 12; i++) 
-        G728_Unpack(((const BYTE *)buffer)+i*5, temp_frame_buffer+i*4);
-      written = 60;
-      break;
-
-    case G729B :
-      writeBuf = temp_frame_buffer;
-      writeLen = 12;
-
-      if (count == 2) {
-        temp_frame_buffer[0] = 2;
-        temp_frame_buffer[1] = *(const WORD *)buffer;
-        memset(&temp_frame_buffer[2], 0, 8);
-        written = 2;
-      }
-      else {
-        if (memcmp(buffer, "\0\0\0\0\0\0\0\0\0", 10) != 0)
-          temp_frame_buffer[0] = 1;
-        else
-          temp_frame_buffer[0] = 0;
-        memcpy(&temp_frame_buffer[1], buffer, 10);
-        written = 10;
-      }
-      break;
-
-    default :
-      writeBuf = buffer;
-      writeLen = writeFrameSize;
+  {
+    PWaitAndSignal rmutex(readMutex);
+  
+    written = 0;
+  
+    if (writeStopped) 
+      return FALSE;
+  
+    if (readStopped) {
+      PThread::Sleep(30);
       written = writeFrameSize;
-  }
-
-  if (count < written) {
-    osError = EINVAL;
-    PTRACE(1, "xJack\tWrite of too small a buffer : " << count << " vs " << written);
-    return FALSE;
-  }
-
-  for (;;) {
-
-    fd_set wfds;
-    FD_ZERO(&wfds);
-    FD_SET(os_handle, &wfds);
-    struct timeval ts;
-    ts.tv_sec = 5;
-    ts.tv_usec = 0;
-    int stat = ::select(os_handle+1, NULL, &wfds, NULL, &ts);
-
-    if (stat == 0) {
-      PTRACE(1, "IXJ\tWrite timeout");
-      return FALSE;
+      return TRUE;
     }
-
-    if (stat > 0) {
-      stat = ::write(os_handle, writeBuf, writeLen);
-      if (stat == (int)writeLen)
+  
+    WORD temp_frame_buffer[48];
+    const void * writeBuf;
+    int writeLen;
+  
+    switch (CodecInfo[writeCodecType].mode) {
+      case G723_63:
+      case G723_53:
+        writeBuf = buffer;
+        writeLen = 24;
+        written = G723count[(*(BYTE *)buffer)&3];
         break;
+  
+      case G728 :
+        writeBuf = temp_frame_buffer;
+        writeLen = sizeof(temp_frame_buffer);
+  
+        // for G728, unpack twelve x 4 vectors
+        PINDEX i;
+        for (i = 0; i < 12; i++) 
+          G728_Unpack(((const BYTE *)buffer)+i*5, temp_frame_buffer+i*4);
+        written = 60;
+        break;
+  
+      case G729B :
+        writeBuf = temp_frame_buffer;
+        writeLen = 12;
+  
+        if (count == 2) {
+          temp_frame_buffer[0] = 2;
+          temp_frame_buffer[1] = *(const WORD *)buffer;
+          memset(&temp_frame_buffer[2], 0, 8);
+          written = 2;
+        }
+        else {
+          if (memcmp(buffer, "\0\0\0\0\0\0\0\0\0", 10) != 0)
+            temp_frame_buffer[0] = 1;
+          else
+            temp_frame_buffer[0] = 0;
+          memcpy(&temp_frame_buffer[1], buffer, 10);
+          written = 10;
+        }
+        break;
+  
+      default :
+        writeBuf = buffer;
+        writeLen = writeFrameSize;
+        written = writeFrameSize;
     }
-
-    if ((stat >= 0) || (errno != EINTR)) {
-      PTRACE(1, "IXJ\tWrite error = " << errno);
+  
+    if (count < written) {
+      osError = EINVAL;
+      PTRACE(1, "xJack\tWrite of too small a buffer : " << count << " vs " << written);
       return FALSE;
     }
+  
+    for (;;) {
+  
+      fd_set wfds;
+      FD_ZERO(&wfds);
+      FD_SET(os_handle, &wfds);
+      struct timeval ts;
+      ts.tv_sec = 5;
+      ts.tv_usec = 0;
+      int stat = ::select(os_handle+1, NULL, &wfds, NULL, &ts);
+  
+      if (stat == 0) {
+        PTRACE(1, "IXJ\tWrite timeout");
+        return FALSE;
+      }
+  
+      if (stat > 0) {
+        stat = ::write(os_handle, writeBuf, writeLen);
+        if (stat == (int)writeLen)
+          break;
+      }
+  
+      if ((stat >= 0) || (errno != EINTR)) {
+        PTRACE(1, "IXJ\tWrite error = " << errno);
+        return FALSE;
+      }
 
-    PTRACE(1, "IXJ\tWrite EINTR");
+      PTRACE(1, "IXJ\tWrite EINTR");
+    }
   }
 
 //  PTRACE(4, "IXJ\tWrote " << writeLen << " bytes to codec");
 
+  PThread::Yield();
+
   return TRUE;
 }
 
@@ -1598,47 +1764,79 @@ unsigned OpalIxJDevice::GetAverageSignalLevel(unsigned, BOOL playback)
 
 BOOL OpalIxJDevice::EnableAudio(unsigned line, BOOL enable)
 {
+  if (line >= GetLineCount())
+    return FALSE;
+
   int port = PORT_SPEAKER;
 
-  if (enable)
+  if (enable) {
+    if (enabledAudioLine != line) {
+      if (enabledAudioLine != UINT_MAX && exclusiveAudioMode) {
+        PTRACE(3, "xJack\tEnableAudio on port when already enabled other port.");
+        return FALSE;
+      }
+      enabledAudioLine = line;
+    }
     port = (line == POTSLine) ? PORT_POTS : PORT_PSTN;
+  }
+  else
+    enabledAudioLine = UINT_MAX;
 
   return ConvertOSError(IOCTL2(os_handle, IXJCTL_PORT, port));
 }
 
 
-static unsigned ConvertVolume(unsigned volume, unsigned max)
+BOOL OpalIxJDevice::IsAudioEnabled(unsigned line)
 {
-  // Use four tier logarithmic scale
-  unsigned zone1 = max/64;
-  if (volume < 25)
-    return volume*zone1/25;
+  return enabledAudioLine == line;
+}
 
-  unsigned zone2 = max/16;
-  if (volume < 50)
-    return (volume-25)*(zone2-zone1)/25 + zone1;
+PINDEX OpalIxJDevice::LogScaleVolume(unsigned line, PINDEX volume, BOOL isPlay)
+{
+  PINDEX dspMax = isPlay ? 0x100 : 0x200;
 
-  unsigned zone3 = max/4;
-  if (volume < 75)
-    return (volume-50)*(zone3-zone2)/25 + zone2;
+  switch (dwCardType) {
+  case 0:
+  case PhoneJACK:
+    dspMax = isPlay ? 0x100 : 0x200;
+    break;
+  case LineJACK:
+    dspMax = 0x200;
+    break;
+  case PhoneJACK_Lite:
+    dspMax = 0x200;
+    break;
+  case PhoneJACK_PCI:
+    dspMax = 0x100;
+    break;
+  case PhoneCARD:
+    dspMax = 0x200;
+    break;    
+  case PhoneJACK_PCI_TJ:
+    if (line == POTSLine)
+      dspMax = 0x100;
+    else
+      dspMax = 0x60;
+  }
 
-  if (volume < 100)
-    return (volume-75)*(max-zone3)/25 + zone3;
+/* The dsp volume is exponential in nature.
+   You can plot this exponential function with gnuplot.
+   gnuplot
+   plot [t=0:100] exp((t/20) -1) / exp(4) */
+  PINDEX res = (PINDEX) (dspMax * exp((((double)volume) / 20.0) - 1) / exp(4.0));
 
-  return max;
-}  
+  return res;
+}
  
 
-BOOL OpalIxJDevice::SetRecordVolume(unsigned, unsigned volume)
+BOOL OpalIxJDevice::SetRecordVolume(unsigned line, unsigned volume)
 {
   PWaitAndSignal mutex1(readMutex);
   userRecVol = volume;
   if ((aecLevel == AECAGC) || inRawMode)
     return TRUE;
 
-  // Need to figure out values for each card type here.
-  unsigned dspRecMax = 0x200;
-  return IOCTL2(os_handle, IXJCTL_REC_VOLUME, ConvertVolume(volume, dspRecMax));
+  return IOCTL2(os_handle, IXJCTL_REC_VOLUME, LogScaleVolume(line, volume, FALSE));
 }
 
 BOOL OpalIxJDevice::GetRecordVolume(unsigned, unsigned & volume)
@@ -1647,16 +1845,14 @@ BOOL OpalIxJDevice::GetRecordVolume(unsigned, unsigned & volume)
   return TRUE;
 }
 
-BOOL OpalIxJDevice::SetPlayVolume(unsigned, unsigned volume)
+BOOL OpalIxJDevice::SetPlayVolume(unsigned line, unsigned volume)
 {
   PWaitAndSignal mutex1(readMutex);
   userPlayVol = volume;
   if (inRawMode)
     return TRUE;
 
-  // Need to figure out values for each card type here.
-  unsigned dspPlayMax = 0x100;
-  return IOCTL2(os_handle, IXJCTL_PLAY_VOLUME, ConvertVolume(volume, dspPlayMax));
+  return IOCTL2(os_handle, IXJCTL_PLAY_VOLUME, LogScaleVolume(line, volume, TRUE));
 }
 
 BOOL OpalIxJDevice::GetPlayVolume(unsigned, unsigned & volume)
@@ -1689,6 +1885,24 @@ BOOL OpalIxJDevice::SetAEC(unsigned line, AECLevels level)
 }
 
 
+unsigned OpalIxJDevice::GetWinkDuration(unsigned)
+{
+  if (!IsOpen())
+    return 0;
+
+  return IOCTL2(os_handle, IXJCTL_WINK_DURATION, 0);
+}
+
+
+BOOL OpalIxJDevice::SetWinkDuration(unsigned, unsigned winkDuration)
+{
+  if (!IsOpen())
+    return FALSE;  
+
+  return IOCTL2(os_handle, IXJCTL_WINK_DURATION, winkDuration);
+}
+
+
 BOOL OpalIxJDevice::GetVAD(unsigned)
 {
   return FALSE;
@@ -1772,9 +1986,9 @@ static void FormatCallerIdString(const PString & idString, PHONE_CID & callerIdI
   }
 
   // truncate name and number fields
-  if (name.GetLength() > sizeof(callerIdInfo.name))
+  if (name.GetLength() > (PINDEX)sizeof(callerIdInfo.name))
     name = name.Left(sizeof(callerIdInfo.name));
-  if (number.GetLength() > sizeof(callerIdInfo.number))
+  if (number.GetLength() > (PINDEX)sizeof(callerIdInfo.number))
     number = number.Left(sizeof(callerIdInfo.number));
 
   sprintf(callerIdInfo.month, "%02i", theTime.GetMonth());
@@ -1873,11 +2087,11 @@ BOOL OpalIxJDevice::PlayDTMF(unsigned, const char * tones, DWORD onTime, DWORD o
 
     IOCTL2(os_handle, PHONE_PLAY_TONE, code);
 
-    PThread::Current()->Sleep(onTime + offTime);
+    PThread::Sleep(onTime + offTime);
 
     long countDown = 200;  // a tone longer than 2 seconds? I don't think so...
     while ((countDown > 0) && IOCTL(os_handle, PHONE_GET_TONE_STATE) != 0) {
-      PThread::Current()->Sleep(10);
+      PThread::Sleep(10);
       countDown--;
     }
     if (countDown == 0)
@@ -2167,9 +2381,9 @@ BOOL OpalIxJDevice::SetCountryCode(T35CountryCodes country)
   }
 
   if (country == UnknownCountry) {
-    PTRACE(4, "IXJ\tRequest to set DAA country to unknown country code " << country);
+    PTRACE(4, "IXJ\tRequest to set DAA country to unknown country code");
   } else {
-    PTRACE(4, "IXJ\tSetting DAA country code to " << country);
+    PTRACE(4, "IXJ\tSetting DAA country code to " << (int)country);
     static int ixjCountry[NumCountryCodes] = {
       DAA_JAPAN, 0, 0, 0, DAA_GERMANY, 0, 0, 0, 0, DAA_AUSTRALIA, 0, 0, 0, 0,
       0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
@@ -2203,7 +2417,7 @@ PStringArray OpalIxJDevice::GetDeviceNames()
   for (i = 0; i < 10; i++) {
     PString devName = psprintf("/dev/phone%i", i);
     int handle = ::open((const char *)devName, O_RDWR);
-    if (handle < 0)
+    if (handle < 0 && errno != EBUSY)
       continue;
     ::close(handle);
     array[j++] = devName;
@@ -2211,5 +2425,7 @@ PStringArray OpalIxJDevice::GetDeviceNames()
   return array;
 }
 
+#endif // HAS_IXJ
+
 
 /////////////////////////////////////////////////////////////////////////////
